package dev.fyloz.musicplayer.core

import com.typesafe.config.ConfigFactory
import dev.fyloz.musicplayer.core.data.RepositoryInjection
import dev.fyloz.musicplayer.core.factory.TrackFactoryProxy
import dev.fyloz.musicplayer.core.http.auth.AuthorizationData
import dev.fyloz.musicplayer.core.http.configureTrackRoutes
import dev.fyloz.musicplayer.core.logic.TrackLogic
import dev.fyloz.musicplayer.modules.Module
import io.ktor.client.*
import io.ktor.http.*
import io.ktor.serialization.kotlinx.json.*
import io.ktor.server.application.*
import io.ktor.server.config.*
import io.ktor.server.engine.*
import io.ktor.server.netty.*
import io.ktor.server.plugins.callloging.*
import io.ktor.server.plugins.contentnegotiation.*
import io.ktor.server.plugins.cors.routing.*
import io.ktor.server.routing.*
import kotlinx.serialization.json.Json
import org.koin.dsl.module
import org.koin.ktor.plugin.Koin
import org.koin.logger.slf4jLogger
import org.slf4j.event.Level

fun main() {
    embeddedServer(Netty, port = 8080, host = "0.0.0.0", module = Application::module)
        .start(wait = true)
}

fun Application.module() {
    val config = HoconApplicationConfig(ConfigFactory.load())

    log.info("Loading modules...")
    registerModules(config)

    install(CallLogging) {
        level = Level.DEBUG
//        filter { call -> call.request.path().startsWith("/") }
    }

    install(ContentNegotiation) {
        json()
    }

    module {
        install(Koin) {
            slf4jLogger()
            modules(
                RepositoryInjection.koinBeans,

                module {
                    single { TrackFactoryProxy() }
                    single { TrackLogic() }
                    single {
                        HttpClient {
                            install(io.ktor.client.plugins.contentnegotiation.ContentNegotiation) {
                                json(Json {
                                    ignoreUnknownKeys = true
                                })
                            }

//                            install(Logging) {
//                                level = LogLevel.ALL
////        filter { call -> call.request.path().startsWith("/") }
//                            }
                        }
                    }

                    registeredModules.values.forEach { it.setupDependencyInjection(this, config) }
                }
            )
        }
    }

    registeredModules.values.forEach { it.configure(this) }

    install(Routing) {
        registeredModules.values.forEach { it.configureRoutes(this) }

        route("/api/v1") {
            configureTrackRoutes()
        }
    }

    install(CORS) {
        allowMethod(HttpMethod.Options)
        allowMethod(HttpMethod.Post)
        allowMethod(HttpMethod.Put)
        allowMethod(HttpMethod.Delete)
        allowHeader(HttpHeaders.AccessControlAllowOrigin)
        allowHeader(HttpHeaders.ContentType)
        anyHost() // @TODO: Don't do this in production if possible. Try to limit it.
        allowCredentials = true
    }
}

private val registeredModules = mutableMapOf<String, Module>()

fun ApplicationCall.getAuthorizationData(): AuthorizationData {
    val auth = AuthorizationData()

    registeredModules.forEach { (moduleName, module) ->
        val data = module.getAuthorizationData(this)
        if (data != null) {
            auth.addModuleData(data, moduleName)
        }
    }

    return auth
}

private fun Application.registerModules(config: ApplicationConfig) {
    val enabledModules = config.property("modules.enabled").getList()
    enabledModules.forEach { registerModule(it, config) }
}

private fun Application.registerModule(moduleName: String, config: ApplicationConfig) {
    log.debug("Found module '$moduleName'")
    val moduleClass = config.property("modules.$moduleName.class").getString()
    val module = loadModule(moduleClass)

    registeredModules[moduleName] = module
    log.info("Registered module '$moduleName'")
}

private fun Application.loadModule(moduleClass: String): Module {
    log.debug("Loading module class '$moduleClass'...")
    return Class.forName(moduleClass).getDeclaredConstructor().newInstance() as Module
}
